/**************************************************************************
	LGBCheckbox
	
	Public domain, by Zig Zichterman.
	
	This class implements 3D checkboxes according to the guidelines
	suggested in _develop_ 15. Some of the drawing code is taken from
	the public domain source accompanying _develop_ 15.

	1.0b5
		11/28/94	zz	use grayishTextOr for inactive color button title
		11/27/94	zz	add Calc/UseBackgroundColor(), mBackgroundColor
						make mControl a ControlHandle (was ControlRecord*)
						use classes for font, pen save/restore
		11/25/94	zz	explicitly setup font for offscreen GWorld
						(fixes useWFont bug)
		10/13/94	zz	change tristate drawing to match Apple's
						standard (see Installer 4.3) instead of 
						Microsoft's (see Word, Excel, and others)
	1.0b4
		07/31/94	zz	Draw 3D if 4-bit grey, BW if 4-bit color
		07/31/94	zz	save/restore pen colors
	1.0b3
		07/28/94	zz	offscreen drawing for less flashy TextBox()
	1.0b2 (never released)
		07/20/94	zz	call PenNormal() before drawing anything!
	1.0b1
**************************************************************************/
#include "LGBCheckbox.h"

#include <GestaltEqu.h>

//#include "LGBColorTable.h"
#include "LGBControl.h"
#include "LGBDeviceIterator.h"
#include "LGBClipRegion.h"
#include "LGBFont.h"
#include "LGBPen.h"
#include "UGBDraw.h"

const short	LGBCheckbox_boxWidth	= 12;

/**************************************************************************
	Main()														[static]
	
	Main entry point for all checkbox calls. Dispatch according to
	message.
**************************************************************************/
long
LGBCheckbox::Main(
	short			inVariation,
	ControlHandle	ioControl,
	short			inMsg,
	long			ioParam)
{
	long			returnMe	= 0;

	// lock the control handle for the duration of this call
	char state = ::HGetState((Handle) ioControl);
	::HLock((Handle) ioControl);
	
	LGBCheckbox	checkbox(ioControl,(inVariation & useWFont)?true:false);
	
	switch (inMsg) {
		case drawCntl	:
				checkbox.Draw(ioParam);
				break;
							
		case testCntl	:
			{
				Point	hitPt;
				hitPt.h	= LoWord(ioParam);
				hitPt.v	= HiWord(ioParam);
				if (checkbox.Test(hitPt)) {
					returnMe = inButton;
				}
			}
			break;
		
		case calcCRgns	:	// only called in 24-bit mode
			{	// is 32-bit addressing off?
				long	result	= 0;
				OSErr	err = ::Gestalt(gestaltAddressingModeAttr,&result);
				if (!err && ((result &
						(1L << gestalt32BitAddressing)) == 0)) {
					RgnHandle	rgn = (RgnHandle)
									::StripAddress((Ptr) ioParam);
					checkbox.CalcCRgn(rgn);
				}
			}
			break;
		
		case calcCntlRgn	:	// only called in 32-bit mode
			checkbox.CalcCRgn((RgnHandle) ioParam);
			break;
	}
	
	// unlock handle
	::HSetState((Handle) ioControl,state);
	return returnMe;
}

//
// Constructor
//

/**************************************************************************
	LGBCheckbox(ControlHandle,Boolean)
	
	construct/initialize a control object. Just stores the control handle
	and the "should I use the window font?" flag in data members.
**************************************************************************/
LGBCheckbox::LGBCheckbox(
	ControlHandle	inControl,
	Boolean			inUseWFont)
	: mControl(inControl),
	  mUseWFont(inUseWFont)
{
	mWindowPort	= NULL;
	
	mBackgroundColor.red	=
	mBackgroundColor.green	=
	mBackgroundColor.blue	= 0;
	mHighlightColor		= mBackgroundColor;
	mShadowColor		= mBackgroundColor;
	mInactiveFrameColor	= mBackgroundColor;
}

//
// Dispatch entry points
//

/**************************************************************************
	Draw()
	
	Draw the control. inPartCode is a part code specifying which part of
	the control to draw, or 0 for the entire control.
	
	Checkboxes only have inButton, so we ignore the inPartCode.

	Save the current drawing environment. Iterate through all the devices
	(screens), drawing the control in color or black and white depending
	on the screen depth. Once done, restore the drawing environment and
	return.
**************************************************************************/
void
LGBCheckbox::Draw(
	long	inPartCode)
{
	// if we're invisible, don't draw
	if ((**mControl).contrlVis == false) return;
	
	// remember the current port--it's our window
	::GetPort(&mWindowPort);
	
	// figure out the correct colors
	CalcColors();
	
	// save the font
	LGBFont	saveFont;

	// save the pen colors, normalize
	LGBPen	savePen;
	LGBPen::Normal();
	
	// save the clip region
	LGBClipRegion	saveClip;
	if (saveClip.HasClip() == false) {
		return;
	}
	
	// make the pen something sensible
	::PenNormal();
		
	// loop through all the devices (screens)
	UGBDraw::Offscreen	offscreen;
	LGBDeviceIterator	device;
	device.Init((**mControl).contrlRect);
	short	depth	= 0;
	do {
		depth = device.Next();
		if (depth == 0) break;	// all done with devices
		UGBDraw::OffscreenPre(offscreen);
		// if we're drawing to an offscreen GWorld, we
		// must force the GWorld's font setup to the
		// onscreen port, otherwise useWFont uses
		// System Font (Chicago), default size (12).
		saveFont.Restore();

		if ((depth < 4)
				|| (depth == 4 && device.mDeviceIsColor)) {
			DrawBW();
		} else {
			DrawColor();
		}
		UGBDraw::OffscreenPost(offscreen);
	} while(true);

	// dump the device iterator
	device.Dispose();
	
	// restore the clip region
	saveClip.Restore();
	saveClip.Dispose();
	
	// restore the pen
	savePen.Restore();

	// restore the font
	saveFont.Restore();
}

/**************************************************************************
	Test()
	
	Return inButton if the point is in our rect
**************************************************************************/
Boolean
LGBCheckbox::Test(
	Point inHitPt)
{
	return ::PtInRect(inHitPt,&((**mControl).contrlRect));
}

/**************************************************************************
	CalcCRgn()
	
	Calculate the control's region in the given region handle.
	The region is just our bounding rect.
**************************************************************************/
void
LGBCheckbox::CalcCRgn(
	RgnHandle ioRgn)
{
	if (!ioRgn) return;		// idiot resistance
	::RectRgn(ioRgn,&((**mControl).contrlRect));
}

//
// Draw
//

/**************************************************************************
	DrawBW()
	
	Draw the control in black and white. It is NOT safe to make any
	color QuickDraw calls here.
**************************************************************************/
void
LGBCheckbox::DrawBW(void)
{
	// Set the pen to black, 1x1
	::PenNormal();

	// set up our font
	if (!mUseWFont) LGBControl::SetupFont();
	
	// calculate box locations
	Rect	checkbox,titleBox;
	CalcBoxes(checkbox,titleBox);
	
	// draw the parts
	DrawCheckbox(checkbox);
	DrawTitle(titleBox, (**mControl).contrlHilite == 255);
}

/**************************************************************************
	DrawColor()
	
	Draw the control in color
**************************************************************************/
void
LGBCheckbox::DrawColor(void)
{
	// draw active or inactive?
	if ((**mControl).contrlHilite == 255) {
		DrawColorInactive();
	} else {
		DrawColorActive();
	}

	// restore the pen
	::PenNormal();
	UGBDraw::PenReallyNormal();
}

/**************************************************************************
	CalcBoxes()
	
	Figure out where to draw the checkbox and the title.
**************************************************************************/
void
LGBCheckbox::CalcBoxes(
	Rect	&outCheckbox,
	Rect	&outTitleBox)
{
	const Rect	contrlRect	= (**mControl).contrlRect;
	const StringPtr	title	= (**mControl).contrlTitle;
	LGBControl::CalcBoxes(contrlRect,title,
					outCheckbox,outTitleBox);
}

/**************************************************************************
	DrawCheckbox()
	
	Draw the checkbox. Draw the 3D effect, the "X", everything
	but the title and the background.
**************************************************************************/
void
LGBCheckbox::DrawCheckbox(
	const Rect	&inCheckbox,
	Boolean		inColor)
{
	// draw the frame
	::FrameRect(&inCheckbox);
	
	// get the value
	const short value = (**mControl).contrlValue;
	
	// erase the insides unless MS-tristated
	if (value <= 2) {			
		Rect	eraseMe;
		eraseMe.top		= inCheckbox.top + 1;
		eraseMe.left	= inCheckbox.left + 1;
		eraseMe.right	= inCheckbox.right - 1;
		eraseMe.bottom	= inCheckbox.bottom - 1;
		::EraseRect(&eraseMe);
	}
	
	// draw the X or erase the guts
	if (value == 0) {			// clear checkbox
		// do nothing
	} else if (value == 1) {	// X
		Rect	x;
		x.left		= inCheckbox.left;
		x.top		= inCheckbox.top;
		x.right		= inCheckbox.right - 1;
		x.bottom	= inCheckbox.bottom - 1;
		::MoveTo(x.left,x.top);
		::LineTo(x.right,x.bottom);
		::MoveTo(x.left,x.bottom);
		::LineTo(x.right,x.top);
	} else if (value == 2) {	// Apple tristate
		// draw a dash through the middle of the box
		const short	centerV	= (inCheckbox.top + inCheckbox.bottom) / 2;
		::BackColor(whiteColor);
		::MoveTo(inCheckbox.left + 3,centerV);
		::LineTo(inCheckbox.right - 4, centerV);
	} else { 	// value == 3+	// Microsoft tristate
		// fabricate a ltGray pattern on the fly. We don't
		// have quickdraw globals and I'm not going to set
		// up A4 to get them.
		unsigned char ltGrey[8];
		ltGrey[0] = ltGrey[2] = ltGrey[4] = ltGrey[6] = 0x88;
		ltGrey[1] = ltGrey[3] = ltGrey[5] = ltGrey[7] = 0x22;
		
		if (inColor) {
			// draw a greyish pattern, not black on white
			UGBDraw::ForeGrey(UGBDraw_grey7);
			UGBDraw::BackGrey(UGBDraw_greyD);
		}
		Rect	grayRect = inCheckbox;
		::InsetRect(&grayRect,1,1);
		::FillRect(&grayRect,(PatPtr) &ltGrey);
		if (inColor) {
			UGBDraw::PenNormal();
		}
	}
	
	// draw highlight
	if (((**mControl).contrlHilite)
			&& ((**mControl).contrlHilite != 255)) {
		::PenNormal();
		Rect	hilite = inCheckbox;
		::InsetRect(&hilite,1,1);
		::FrameRect(&hilite);
	}
}

/**************************************************************************
	DrawTitle()
	
	Draw the title.
**************************************************************************/
void
LGBCheckbox::DrawTitle(
	const Rect	&inTitleBox,
	Boolean		inDim1Bit)
{
	const StringPtr	title = (**mControl).contrlTitle;
	::TextBox(title + 1,*title,&inTitleBox,teFlushDefault);
	
	// grey out the name if inactive
	if (inDim1Bit) {
		unsigned char grey[8];
		grey[0] = grey[2] = grey[4] = grey[6] = 0xAA;
		grey[1] = grey[3] = grey[5] = grey[7] = 0x55;
		::PenPat((PatPtr) grey);
		::PenMode(patBic);
		::PaintRect(&inTitleBox);
		::PenNormal();
	}
}

/**************************************************************************
	DrawColorInactive()
	
	Draw the checkbox, in color, but flattened out and inactive
**************************************************************************/
void
LGBCheckbox::DrawColorInactive(void)
{
	// Set the pen to black, 1x1
	::PenNormal();
	
	// set up our font
	if (!mUseWFont) LGBControl::SetupFont();
	
	// calculate box locations
	Rect	checkbox,titleBox;
	CalcBoxes(checkbox,titleBox);
	
	// grey out our entire background
	UseBackgroundColor();
	::EraseRect(&((**mControl).contrlRect));
	
	// set the pen to the "inactive control" color
	// leave the backcolor as background color
	UseInactiveFrameColor();	
	DrawCheckbox(checkbox,true);
	
	// draw the button title
	// set the background to grey
	DrawTitle(titleBox);
}

/**************************************************************************
	DrawColorActive()
	
	Draw the checkbox, in 3D color
**************************************************************************/
void
LGBCheckbox::DrawColorActive(void)
{
	// Set the pen to black, 1x1
	::PenNormal();
	
	// set up our font
	if (!mUseWFont) LGBControl::SetupFont();
	
	// calculate box locations
	Rect	checkbox,titleBox;
	CalcBoxes(checkbox,titleBox);
	
	// assume we're erasing to a normal window
	UseBackgroundColor();

	// grey out our entire background
	::EraseRect(&((**mControl).contrlRect));
	
	// set the pen to the "active control" color
	// set the backcolor to white
	UGBDraw::PenFrameActive();
	UGBDraw::BackGrey(UGBDraw_greyF);
	DrawCheckbox(checkbox,true);
	
	// draw the button title
	// set the background to grey
	UseBackgroundColor();
	DrawTitle(titleBox, (**mControl).contrlHilite == 255);
	
	// draw 3D effect around checkbox
		// top left sides
	UseShadowColor();
	::MoveTo(checkbox.left - 1,checkbox.bottom);
	::LineTo(checkbox.left - 1,checkbox.top - 1);
	::LineTo(checkbox.right,checkbox.top - 1);
	
		// bottom right sides
	UseHighlightColor();
	::MoveTo(checkbox.left,checkbox.bottom);
	::LineTo(checkbox.right,checkbox.bottom);
	::LineTo(checkbox.right,checkbox.top);
}

/**************************************************************************
	CalcColors()									[protected]
	
	Figure out the appropriate colors to use for our color drawing.
	
	Background--this is usually light grey (UGBDraw_background == 0xDDDD),
	but if the window has a custom color for the content region, use that.
	If the control has a custom color for the body region, use that.
	
	Highlight--a blend of white and the background color
	
	Shadow--a blend of black and the background color
**************************************************************************/
void
LGBCheckbox::CalcColors(void)
{
	LGBControl::CalcBackgroundColor(mControl,mWindowPort,mBackgroundColor);
	
	// blend white and background for highlight
	UGBDraw::Blend(UGBDraw_white,mBackgroundColor,0.50,mHighlightColor);
	
	// blend black and background for shadow
	UGBDraw::Blend(UGBDraw_black,mBackgroundColor,0.30,mShadowColor);
	
	// blend black and background for inactive text/frame
	UGBDraw::Blend(UGBDraw_black,mBackgroundColor,0.45,mInactiveFrameColor);
}

/**************************************************************************
	UseBackgroundColor()									[protected]
	
	Set the background color to our previously-calculated color
**************************************************************************/
void
LGBCheckbox::UseBackgroundColor(void)
{
	::RGBBackColor(&mBackgroundColor);
}

/**************************************************************************
	UseHighlightColor()										[protected]
	
	Set the foreground color to the highlight color.
**************************************************************************/
void
LGBCheckbox::UseHighlightColor(void)
{
	::RGBForeColor(&mHighlightColor);
}

/**************************************************************************
	UseShadowColor()										[protected]
	
	Set the foreground color to the shadow color.
**************************************************************************/
void
LGBCheckbox::UseShadowColor(void)
{
	::RGBForeColor(&mShadowColor);
}

/**************************************************************************
	UseInactiveFrameColor()										[protected]
	
	Set the foreground color to the inactive frame color.
**************************************************************************/
void
LGBCheckbox::UseInactiveFrameColor(void)
{
	::RGBForeColor(&mInactiveFrameColor);
}
