/*
	Public domain, by Zig Zichterman.
*/
/*
	LGBClipRgn
	
	Storage for the clip region.  Use this class to save and
	restore the clip region across draws
	
	avoid destructor tricks to avoid requiring C++ runtime stuff.
	
	11/27/94	zz	h	initial write
*/
#include "LGBClipRegion.h"

/**************************************************************************
	LGBClipRegion(void)									[public]
	
	Create storage for the current clip region and copy the current
	clip region into it. On error, do nothing.
**************************************************************************/
LGBClipRegion::LGBClipRegion(void)
{
	mClipRgn = NULL;
	Save();
}

/**************************************************************************
	Dispose()											[public]
	
	Because we cannot have a destructor (without incurring the wratch
	of C++ runtime, which makes this CDEF overly complex), we politely
	ask that our callers explicitly call Dispose()
**************************************************************************/
void
LGBClipRegion::Dispose(void)
{
	// if we're using any storage
	if (mClipRgn) {
		// dump it
		::DisposeRgn(mClipRgn);
		mClipRgn = NULL;
	}
}

/**************************************************************************
	Save()												[public]
	
	Save the current clip region to our local region. Discards any
	previously saved clip
**************************************************************************/
void
LGBClipRegion::Save(void)
{
	// if no region, make one
	if (mClipRgn == NULL) {
		mClipRgn = ::NewRgn();
	}
	
	// if STILL no region, error. Silently abort
	if (mClipRgn == NULL) {
		return;
	}
	
	// save the current clip
	::GetClip(mClipRgn);
}

/**************************************************************************
	Restore()											[public]
	
	Set the current clip to whatever we've most recently saved. Do 
	nothing if never saved
**************************************************************************/
void
LGBClipRegion::Restore(void) const
{
	// if never saved
	if (mClipRgn == NULL) {
		// do nothing
		return;
	}
	
	// restore clip region
	::SetClip(mClipRgn);
}

/**************************************************************************
	HasClip()											[public]
	
	Do we have a saved clip region?
**************************************************************************/
Boolean
LGBClipRegion::HasClip(void) const
{
	return (mClipRgn != NULL);
}


