/**************************************************************************
	LGBControl
	
	Public domain, by Zig Zichterman.
	
	Just a few utilities that the control classes use.
	
	Some of the drawing code is taken from the public domain source
	accompanying _develop_ 15.

	1.0b5
		11/27/94	zz	h	add CalcBackgroundColor()
**************************************************************************/
#include "LGBControl.h"

#include "LGBColorTable.h"
#include "UGBDraw.h"

const LGBControl_boxWidth	= 12;	// width of checkboxes, radio buttons


//
// Drawing Utilities
//

/**************************************************************************
	CountLines()										[public, static]
	
	Count the number of carriage return ('\r') characters in the
	given string.
	
	carriage return is NEVER part of a 2-byte character, so this will
	work with double byte scripts, too.
**************************************************************************/
short
LGBControl::CountLines(
	const StringPtr inString)
{
	// start with one line
	short			numLines		= 1;
	
	// iterate through all the characters
	short			charsRemaining	= *inString;
	unsigned char	*p				= inString + 1;
	for (;charsRemaining--;p++) {
		// each carriage return implies another line
		if (*p == '\r') numLines++;
	}
	
	// return the number of lines we counted
	return numLines;
}

/**************************************************************************
	TitleHeight()										[public, static]
	
	Return the height of the title in pixels. The font must already be
	set up.
**************************************************************************/
short
LGBControl::TitleHeight(
	const StringPtr inTitle)
{
	// how tall is one line (in pixels?)
	FontInfo	info;
	::GetFontInfo(&info);
	const short	lineHeight	= info.ascent + info.descent + info.leading;
	
	// how many lines are in the title?
	const short	numLines	= CountLines(inTitle);
	
	// return the product of (pixels per line) times (lines)
	return lineHeight * numLines;
}

/**************************************************************************
	CalcBoxes()											[public, static]
	
	Calculate the boxes for the checkbox/radio button and its title
**************************************************************************/
void
LGBControl::CalcBoxes(
	const Rect		&inControlRect,	// frame of control
	const StringPtr	inTitle,		// title of control
	Rect			&outCheckbox,	// where to draw checkbox/radio button
	Rect			&outTitleBox)	// where to draw control title	
{
	//  vertical coords 
	// get the vertical center of the control
	short		centerV		= (inControlRect.top + inControlRect.bottom)/2;
	
	// checkbox is v-centered, 12-px tall
	outCheckbox.top		= centerV - (LGBControl_boxWidth/2);
	outCheckbox.bottom	= outCheckbox.top + LGBControl_boxWidth;

	// textbox is v-centered, tall enough for all lines
	const short		titleHeight	= LGBControl::TitleHeight(inTitle);
	outTitleBox.top				= centerV - (titleHeight/2);
	outTitleBox.bottom			= outTitleBox.top + titleHeight;

	//  horizontal coords 
	// is the Macintosh's current system script predominantly left-to-right?
	// (such as English)
	Boolean		isLeftToRight	= (GetSysJust() == 0);
	
	// checkbox is left- or right-aligned, offset by a couple pix
	if (isLeftToRight) {
		outCheckbox.left	= inControlRect.left + 2;
		outCheckbox.right	= outCheckbox.left + LGBControl_boxWidth;
	} else {
		outCheckbox.right	= inControlRect.right - 2;
		outCheckbox.left	= outCheckbox.right - LGBControl_boxWidth;
	}
	
	// textbox is next to the checkbox
	if (isLeftToRight) {
		outTitleBox.left	= outCheckbox.right + 3;
		outTitleBox.right	= inControlRect.right;
	} else {
		outTitleBox.right	= outCheckbox.left - 3;
		outTitleBox.left	= inControlRect.left;
	}
}

/**************************************************************************
	CalcBackgroundColor()								[static]
	
	Figure out the appropriate color to use for our background. This
	is usually a light grey (UGBDraw_background == 0xDDDD), but if the
	window has a custom color for the content region, use that. If the
	control has a custom color for the body region, use that.
**************************************************************************/
void
LGBControl::CalcBackgroundColor(
	const ControlHandle	inControl,
	const WindowPtr		inWindow,
	RGBColor			&outColor)
{
	// do we have a custom color for our control body?
	{
		LGBColorTable	controlColors;
		controlColors.GetControlColors(inControl);
		// if have color for body
		if (controlColors.GetColorForPart(
				LGBColorTable::Part_ControlBody,outColor)) {
			// look no further
			return;
		}
	}
	
	// do we have a custom color for our window content?
	{
		LGBColorTable	windowColors;
		windowColors.GetWindowColors(inWindow);
		// if have color for content
		if (windowColors.GetColorForPart(
				LGBColorTable::Part_WindowContent,outColor)) {
			// look no further
			return;
		}
	}
	
	// no custom colors. Use default grey
	outColor.red	=
	outColor.green	=
	outColor.blue	= UGBDraw_background;
}

