/**
	SuperSort
	
	A simple (fat) application that demonstrates plug in modules.  Modules can
	be written either in 68k or PowerPC code.  SuperSort looks for a folder named
	"Algorithms" in it's current directory and registers any file of type 'SORT'
	by sticking a UniversalProcPtr into a global array of sorting UniversalProcPtrs.
	If SuperSort is running native on PowerPC, we call the UPP via MixedMode.  If
	we are on 68k, we can call the sorting UPP in the usual way by dereferencing
	the pointer and passing the params directly,
	
	SORT files contain a particular sorting algorithm located in a 'SORT' 128
	resource.  The entry point to a SORT routine is as follows:
	
		void MySortProc(DataPoint *data,short nPoints,SwapProc swap);
	
		where:
			<data> = the data to sort
			<nPoints> = the size of the data set
			<swap> = call back to SuperSort swapping routine

	 1993 Apple Computer, Inc.

	Written by Tim Nichols
	
	Change history:
	
	8/11/93		TCN		New today.
	
**/
#include <Windows.h>
#include <Dialogs.h>
#include <Resources.h>
#include <Menus.h>
#include <Events.h>
#include <OSEvents.h>
#include <OSUtils.h>
#include <Quickdraw.h>
#include <Fonts.h>
#include <Files.h>
#include <Memory.h>
#include <Strings.h>
#include <Packages.h>
#include <Desk.h>
#if powerc
#include <FragLoad.h>
#include <MixedMode.h>
#endif
#include "SuperSort.h"
#include "Timing.h"

/*
	global data
*/

Boolean		gDone = false;		// true when time to quit
Boolean 	gAnimate = true;	// animation on/off?
DataPoint	gData[kLarge];		// the data points to sort
short 		gSize = kSmall;		// current size of data set
long 		gTime = 0;			// time taken to sort	
WindowPtr 	gSortWindow = nil;	// our main window
short		gAlgorithm=1;		// which algorithm is being used
SortProc	gProcs[10];			// UniversalProcPtrs to the sorting routines

#if powerc
ProcInfoType		gSortProcInfo = kCStackBased 	| STACK_ROUTINE_PARAMETER(1,kFourByteCode)
								 					| STACK_ROUTINE_PARAMETER(2,kTwoByteCode)
								 					| STACK_ROUTINE_PARAMETER(3,kFourByteCode);
													
ProcInfoType		gSwapProcInfo =  kCStackBased 	| STACK_ROUTINE_PARAMETER(1,kFourByteCode)
													| STACK_ROUTINE_PARAMETER(2,kFourByteCode);
													
QDGlobals	qd;					// PowerPC uses real live quickdraw globals
#endif


void main(void)
{
#if powerc
	UniversalProcPtr BubbleSortUPP = NewRoutineDescriptor(BubbleSort,gSortProcInfo,kPowerPCISA);
	UniversalProcPtr DoQuickSortUPP = NewRoutineDescriptor(DoQuickSort,gSortProcInfo,kPowerPCISA);
	gProcs[0] = BubbleSortUPP;
	gProcs[1] = DoQuickSortUPP;
#else
	gProcs[0] = BubbleSort;
	gProcs[1] = DoQuickSort;
#endif
	InitToolbox();								// initialze Macintosh toolbox

	MakeMenus();								// create our menus
	
	AddPlugIns();
	
	gSortWindow = GetNewWindow(128,nil,(WindowPtr) -1L );	// create our window
	BailIf(gSortWindow == nil,"\pNo Window");
	ShowWindow(gSortWindow);
	SetPort(gSortWindow);

	InitData();									// initialize the data
	qd.randSeed = (unsigned long)100;			// seed random number generator

	while(!gDone)								// main event loop
		EventLoop();
	
	DisposeWindow(gSortWindow);
}

/*
	initialize the mac toolbox
*/
void InitToolbox(void)
{
	InitGraf(&qd.thePort);
	InitFonts();
	FlushEvents( everyEvent, 0 );
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(0L);
	InitCursor();
}

/*
	set up our menus
*/
void MakeMenus(void)
{
	Handle menuBar;
	
	/*
		get the MBAR resource and put it in the menu bar
	*/
	menuBar = GetNewMBar(128);
	BailIf(menuBar == nil,"\pError getting menu bar");
	SetMenuBar(menuBar);
	DisposeHandle(menuBar);
	
	AddResMenu(GetMHandle(mApple),'DRVR');	// add Apple Menu Items
	
	DrawMenuBar();
}

/*
	look for any plug in modules and add them to the Algorithm menu
	
	plug ins are located in a folder called "Algorithms" in the current directory
	and are of type 'SORT'.
*/
void AddPlugIns(void)
{
	HFileInfo pb;
	OSErr err;
	short vRefNum;
	Str255 name = "\pAlgorithms";
	long dirID;
	MenuHandle menu;
	short i = 2;
	Handle h;
	short refNum;
#if powerc
	FSSpec theFile;
	Boolean foundFrag;
	ConnectionID connID;
	Str255 errName;
	Ptr entry;
	UniversalProcPtr entryUPP;
#endif
	
	err = GetVol(nil,&vRefNum);
	pb.ioCompletion = nil;
	pb.ioVRefNum = vRefNum;
	pb.ioNamePtr = name;
	pb.ioDirID = 0;
	pb.ioFDirIndex = 0;
	err = PBGetCatInfo((CInfoPBPtr)&pb,false);
	BailIf(err != noErr,"\pAlgorithms folder not found.");

	menu = GetMHandle(mAlgorithm);

	/*
		search the Algorithms directory for any
		algorithm files
	*/
	dirID = pb.ioDirID;
	pb.ioFDirIndex = 1;
	do {
		err = PBGetCatInfo((CInfoPBPtr)&pb,false);
		if(err == noErr)
		{
		#if powerc
			/*
				if we're on PowerPC, call CFM to see if there is some native code
				in the data fork of the file.  CFM will return a pointer to the
				main entry point if a fragment is found.  If not, check if there is 
				a 'SORT' resource.
			*/
			foundFrag = false;
			if(FSMakeFSSpec(pb.ioVRefNum,pb.ioFlParID,pb.ioNamePtr,&theFile) == noErr)
			{
				if(GetDiskFragment(&theFile,0,0,theFile.name,kLoadNewCopy,&connID,&entry,errName)==noErr)
				{
					//	we need to make a UPP for this PowerPC plug in for the global sort proc array
					entryUPP = NewRoutineDescriptor(entry,gSortProcInfo,kPowerPCISA);
					gProcs[i++] = entryUPP;
					AppendMenu(menu,pb.ioNamePtr);
					foundFrag = true;
				}
			}
			if(!foundFrag)
		#endif
			if(pb.ioFlFndrInfo.fdType == 'SORT')
			{
				/* 
					open the resource file and get
					the sort resource
				*/
				refNum = HOpenResFile(pb.ioVRefNum,pb.ioFlParID,pb.ioNamePtr,fsRdPerm);
				if(refNum)
				{
					h = GetResource('SORT',128);
					if(h)
					{
						DetachResource(h);
						HLock(h);
						gProcs[i++] = (SortProc)*h;		// remember that a 68k ProcPtr is a UPP
						AppendMenu(menu,pb.ioNamePtr);
					}
					CloseResFile(refNum);
				}
			}
		}
		pb.ioDirID = dirID;			// reset directory id
		pb.ioFDirIndex++;			// look for another plug in
	} while (err == noErr);
	DrawMenuBar();
}

/*
	bail if any problems occur
*/
void BailIf(Boolean b,Str255 s)
{
	if(b)
	{
		ParamText(s,"\p","\p","\p");
		StopAlert(200,nil);
		if(gSortWindow != nil)
			DisposeWindow(gSortWindow);
		ExitToShell();
	}
}

void EventLoop(void)
{
	EventRecord e;
	
	if(WaitNextEvent(everyEvent,&e,5L,nil))
		DoEvent(&e);
}

void DoEvent(EventRecord *e)
{
	switch(e->what)
	{
		case mouseDown:
			DoMouseDown(e);
			break;
		case keyDown:
		case autoKey:
			DoKeyDown(e);
			break;
		case updateEvt:
			DoUpdate((WindowPtr)e->message);
			break;
	}
}

void DoMouseDown(EventRecord *e)
{
	short	part;
	WindowPtr	w;
	
	part = FindWindow(e->where,&w);
	switch(part)
	{
		case inMenuBar:
			AdjustMenus();
			DoMenuCommand(MenuSelect(e->where));
			break;
		case inSysWindow:
			SystemClick(e,w);
			break;
		case inContent:
			if(w!=FrontWindow())
				SelectWindow(w);
			break;
		case inDrag:
			DragWindow(w,e->where,&(**GetGrayRgn()).rgnBBox);
			break;
	}
}

void DoKeyDown(EventRecord *e)
{
	char key;
	
	key = e->message & charCodeMask;
	
	if( (e->modifiers & cmdKey) != 0 )
	{
		AdjustMenus();
		DoMenuCommand(MenuKey(key));
	}
}

void DoUpdate(WindowPtr w)
{
	BeginUpdate(w);
	EraseRect(&qd.thePort->portRect);
	DrawData();
	DrawAlgorithm();
	EndUpdate(w);
}

void DoMenuCommand(long menuResult)
{
	short id,item;
	
	item = menuResult & 0xFFFF;			// low word
	id = (menuResult >> 16) & 0xFFFF;	// high word
	
	switch(id)
	{
		case mApple:
			HandleAppleMenu(item);
			break;
		case mFile:
			HandleFileMenu(item);
			break;
		case mData:
			HandleDataMenu(item);
			break;
		case mAlgorithm:
			HandleAlgorithmMenu(item);
			break;
	}
	HiliteMenu(0);
}

void HandleAppleMenu(short item)
{
	Str255 name;
	DialogPtr d;
	
	switch(item)
	{
		case 1:
			d = GetNewDialog(128,nil,(WindowPtr)-1);
			DrawDialog(d);
			while(!Button());
			DisposDialog(d);
			break;
		default:
			GetItem(GetMHandle(mApple),item,name);
			OpenDeskAcc(name);
			break;
	}

}

void HandleFileMenu(short item)
{
	switch(item)
	{
		case 1:
			gDone = true;
			break;
	}
}

void HandleDataMenu(short item)
{
	switch(item)
	{
		case 1:							// New Data
			NewData();
			break;
		case 2:							// Random Seed...
			GetSeed();
			break;
		case 4:							// Tiny Data Set
			gSize = kSmall;
			NewData();
			break;
		case 5: 						// Medium Data Set
			gSize = kMedium;
			NewData();
			break;
		case 6: 						// Large Data Set
			gSize = kLarge;
			NewData();
			break;
		case 8:							// Animate
			gAnimate = !gAnimate;
			break;
		case 9:							// Sort Data
			Sort(gAlgorithm);
			break;
	}
}

void HandleAlgorithmMenu(short item)
{
	gAlgorithm = item;
	DrawAlgorithm();
}

void AdjustMenus(void)
{
	switch(gSize)
	{	
		case kSmall:
			CheckItem(GetMHandle(mData),4,true);
			CheckItem(GetMHandle(mData),5,false);
			CheckItem(GetMHandle(mData),6,false);
			break;
		case kMedium:
			CheckItem(GetMHandle(mData),4,false);
			CheckItem(GetMHandle(mData),5,true);
			CheckItem(GetMHandle(mData),6,false);
			break;
		case kLarge:
			CheckItem(GetMHandle(mData),4,false);
			CheckItem(GetMHandle(mData),5,false);
			CheckItem(GetMHandle(mData),6,true);
			break;
	}
	CheckItem(GetMHandle(mData),8,gAnimate);
}

/*
	Initialize the data set with new data
*/
void InitData(void)
{
	short i;
	unsigned n;
	long range = kMax - kMin;
	short left = kLeft;
	short height;
	short barWidth;
	
	barWidth = kWidth/gSize - 1;		// size of data set determines width of rect
	for(i=0;i<gSize;i++)				// generate random height for each rect		
	{
		n = Random();					
		n = (n < 0 ? -n : n);			// no negative heights!
		n = n % range;					// chop it to below maximum height
		height = n + kMin;				// make sure it is above minimum height 
		gData[i].n = height;			
		SetRect(&gData[i].r,left,kBottom - height*2,left + barWidth,kBottom);
		left += barWidth + 1;			// move pen 1 to the right of last bar
	}
}

/*
	Create a new data set and redraw it by generating an update event for
	the window.
*/
void NewData(void)
{
	InitData();
	InvalRect(&qd.thePort->portRect);
}

/*
	Draw the data set as a series of rectangles representing the corresponding
	value of the data point.
*/
void DrawData(void)
{
	short	i;
	short 	x;
	
	x = kLeft;
	for(i=0;i<gSize;i++)
	{
		FillRect(&gData[i].r,(ConstPatternParam)qd.black.pat);
	}
}

/*
	set the random number generator seed
*/
void GetSeed(void)
{
	DialogPtr d;
	Handle h;
	Rect r;
	short type;
	short hit;
	Str255 s;
	
	d = GetNewDialog(129,nil,(WindowPtr)-1);
	GetDItem(d,4,&type,&h,&r);
	SelIText(d,4,0,100);
	do{
		ModalDialog(nil,&hit);
	} while((hit != 1) && (hit!=2));
	if(hit == 1)
	{
		GetIText(h,s);
		StringToNum((char *)P2CStr(s),&qd.randSeed);
		NewData();
	}
	DisposeDialog(d);
}

void DrawTime(void)
{
	Rect r;
	char s[256];
	short left,top,right,bottom;
	
	top = qd.thePort->portRect.top + 10;
	left = qd.thePort->portRect.left + 50;
	bottom = top + 70;
	right = left+150;
	SetRect(&r,left,top,right,bottom);
	EraseRect(&r);
	
	MoveTo(left,bottom - 10);
	DrawString("\pSort time : ");
	NumToString(gTime,s);
	DrawString(s);
	DrawString("\p ms");
}

void DrawAlgorithm(void)
{
	Rect r;
	Str255 s;
	short left,top,right,bottom;
	
	left = qd.thePort->portRect.left+210;
	bottom = qd.thePort->portRect.bottom;
	top = bottom - 20;
	right = left + 200;
	SetRect(&r,left,top,right,bottom);
	EraseRect(&r);
	MoveTo(left + 5,bottom - 10);
	DrawString("\pAlgorithm : ");
	GetItem(GetMHandle(mAlgorithm),gAlgorithm,s);
	DrawString(s);
}

void Sort(short which)
{
	long start=0;
	long stop=0;
#if powerc
	UniversalProcPtr 	SwapUPP = NewRoutineDescriptor(Swap,gSwapProcInfo,kPowerPCISA);
#endif
	
//	start = StartTimer();
	
	/*
		if we're on PowerPC, we may be calling either 68k or PowerPC 
		plug in, so we must use CallUniversalProc so Mixed Mode can 
		make a mode switch if necessary.
	*/
	
#if powerc
	CallUniversalProc((UniversalProcPtr)gProcs[which-1],gSortProcInfo,gData,gSize,SwapUPP);
#else
	(*gProcs[which-1])(gData,gSize,Swap);
#endif
	
//	stop = StopTimer();
	if(!gAnimate)
		InvalRect(&qd.thePort->portRect);
	gTime = stop - start;
	DrawTime();
}

void DoQuickSort(DataPoint *data,short size,SwapProc swap)
{
	QuickSort(data,0,size-1,swap);
}

void BubbleSort(DataPoint *data,short size,SwapProc swap)
{
	short i,j;
	
	for(j=0;j<size-1;j++)
	{
		for(i=size-1;i>j;i--)
		{
			if(data[i].n < data[i-1].n)
			{
			#if powerc
				// must use CallUniversalProc since we will be passed a UPP for the SwapProc
				CallUniversalProc(swap,gSwapProcInfo,&data[i],&data[i-1]);
			#else
				(*swap)(&data[i],&data[i-1]);
			#endif
			}
		}
	}
}

short FindPivot(DataPoint *data,short i,short j)
{
	short n;
	short k;
	
	n = data[i].n;
	for(k=i+1;k<=j;k++)
	{
		if(data[k].n > n)
			return k;
		else if(data[k].n < n)
			return i;
	}
	return -1;
}

short Partition(DataPoint *data,short i,short j,short pivot,SwapProc swap)
{
	short l,r;
	
	l = i;
	r = j;
	do {
	#if powerc
		// must use CallUniversalProc since we will be passed a UPP for the SwapProc
		CallUniversalProc(swap,gSwapProcInfo,&data[l],&data[r]);
	#else
		(*swap)(&data[l],&data[r]);
	#endif
		while(data[l].n < pivot)
			l++;
		while(data[r].n >= pivot)
			r--;
	} while(l<=r);
	return l;
}

void QuickSort(DataPoint *data,short first,short last,SwapProc swap)
{
	short pivot;
	short index;
	short k;
	
	index = FindPivot(data,first,last);
	if(index != -1)
	{
		pivot = data[index].n;
		k = Partition(data,first,last,pivot,swap);
		QuickSort(data,first,k-1,swap);
		QuickSort(data,k,last,swap);
	}
}

void Swap(DataPoint *a,DataPoint *b)
{
	short temp;
	
	if(gAnimate)
	{
		// redraw rects
		FillRect(&a->r,(ConstPatternParam)qd.white.pat);
		FillRect(&b->r,(ConstPatternParam)qd.white.pat);
	}
	
	// swap values
	temp = a->n;
	a->n = b->n;
	b->n = temp;
	
	// swap rect tops (other dimensions don't change);
	temp = a->r.top;
	a->r.top = b->r.top;
	b->r.top = temp;
	
	if(gAnimate)
	{
		// redraw rects
		FillRect(&a->r,(ConstPatternParam)qd.black.pat);
		FillRect(&b->r,(ConstPatternParam)qd.black.pat);
	}
}

